package tangle

import (
	"sync"
	"time"

	"github.com/iotaledger/goshimmer/packages/bitutils"
	"github.com/iotaledger/goshimmer/packages/errors"
	"github.com/iotaledger/goshimmer/packages/ternary"
	"github.com/iotaledger/goshimmer/packages/typeutils"
)

// region type definition and constructor //////////////////////////////////////////////////////////////////////////////

type TransactionMetadata struct {
	hash              ternary.Trinary
	hashMutex         sync.RWMutex
	receivedTime      time.Time
	receivedTimeMutex sync.RWMutex
	solid             bool
	solidMutex        sync.RWMutex
	liked             bool
	likedMutex        sync.RWMutex
	finalized         bool
	finalizedMutex    sync.RWMutex
	modified          bool
	modifiedMutex     sync.RWMutex
}

func NewTransactionMetadata(hash ternary.Trinary) *TransactionMetadata {
	return &TransactionMetadata{
		hash:         hash,
		receivedTime: time.Now(),
		solid:        false,
		liked:        false,
		finalized:    false,
		modified:     true,
	}
}

// endregion ///////////////////////////////////////////////////////////////////////////////////////////////////////////

// region getters and setters //////////////////////////////////////////////////////////////////////////////////////////

func (metadata *TransactionMetadata) GetHash() ternary.Trinary {
	metadata.hashMutex.RLock()
	defer metadata.hashMutex.RUnlock()

	return metadata.hash
}

func (metadata *TransactionMetadata) SetHash(hash ternary.Trinary) {
	metadata.hashMutex.RLock()
	if metadata.hash != hash {
		metadata.hashMutex.RUnlock()
		metadata.hashMutex.Lock()
		defer metadata.hashMutex.Unlock()
		if metadata.hash != hash {
			metadata.hash = hash

			metadata.SetModified(true)
		}
	} else {
		metadata.hashMutex.RUnlock()
	}
}

func (metadata *TransactionMetadata) GetReceivedTime() time.Time {
	metadata.receivedTimeMutex.RLock()
	defer metadata.receivedTimeMutex.RUnlock()

	return metadata.receivedTime
}

func (metadata *TransactionMetadata) SetReceivedTime(receivedTime time.Time) {
	metadata.receivedTimeMutex.RLock()
	if metadata.receivedTime != receivedTime {
		metadata.receivedTimeMutex.RUnlock()
		metadata.receivedTimeMutex.Lock()
		defer metadata.receivedTimeMutex.Unlock()
		if metadata.receivedTime != receivedTime {
			metadata.receivedTime = receivedTime

			metadata.SetModified(true)
		}
	} else {
		metadata.receivedTimeMutex.RUnlock()
	}
}

func (metadata *TransactionMetadata) GetSolid() bool {
	metadata.solidMutex.RLock()
	defer metadata.solidMutex.RUnlock()

	return metadata.solid
}

func (metadata *TransactionMetadata) SetSolid(solid bool) bool {
	metadata.solidMutex.RLock()
	if metadata.solid != solid {
		metadata.solidMutex.RUnlock()
		metadata.solidMutex.Lock()
		defer metadata.solidMutex.Unlock()
		if metadata.solid != solid {
			metadata.solid = solid

			metadata.SetModified(true)

			return true
		}
	} else {
		metadata.solidMutex.RUnlock()
	}

	return false
}

func (metadata *TransactionMetadata) GetLiked() bool {
	metadata.likedMutex.RLock()
	defer metadata.likedMutex.RUnlock()

	return metadata.liked
}

func (metadata *TransactionMetadata) SetLiked(liked bool) {
	metadata.likedMutex.RLock()
	if metadata.liked != liked {
		metadata.likedMutex.RUnlock()
		metadata.likedMutex.Lock()
		defer metadata.likedMutex.Unlock()
		if metadata.liked != liked {
			metadata.liked = liked

			metadata.SetModified(true)
		}
	} else {
		metadata.likedMutex.RUnlock()
	}
}

func (metadata *TransactionMetadata) GetFinalized() bool {
	metadata.finalizedMutex.RLock()
	defer metadata.finalizedMutex.RUnlock()

	return metadata.finalized
}

func (metadata *TransactionMetadata) SetFinalized(finalized bool) {
	metadata.finalizedMutex.RLock()
	if metadata.finalized != finalized {
		metadata.finalizedMutex.RUnlock()
		metadata.finalizedMutex.Lock()
		defer metadata.finalizedMutex.Unlock()
		if metadata.finalized != finalized {
			metadata.finalized = finalized

			metadata.SetModified(true)
		}
	} else {
		metadata.finalizedMutex.RUnlock()
	}
}

// returns true if the transaction contains unsaved changes (supports concurrency)
func (metadata *TransactionMetadata) GetModified() bool {
	metadata.modifiedMutex.RLock()
	defer metadata.modifiedMutex.RUnlock()

	return metadata.modified
}

// sets the modified flag which controls if a transaction is going to be saved (supports concurrency)
func (metadata *TransactionMetadata) SetModified(modified bool) {
	metadata.modifiedMutex.Lock()
	defer metadata.modifiedMutex.Unlock()

	metadata.modified = modified
}

// endregion ///////////////////////////////////////////////////////////////////////////////////////////////////////////

// region marshaling functions ////////////////////////////////////////////////////////////////////////////////////////

func (metadata *TransactionMetadata) Marshal() ([]byte, errors.IdentifiableError) {
	marshaledMetadata := make([]byte, MARSHALED_TOTAL_SIZE)

	metadata.receivedTimeMutex.RLock()
	defer metadata.receivedTimeMutex.RUnlock()
	metadata.solidMutex.RLock()
	defer metadata.solidMutex.RUnlock()
	metadata.likedMutex.RLock()
	defer metadata.likedMutex.RUnlock()
	metadata.finalizedMutex.RLock()
	defer metadata.finalizedMutex.RUnlock()

	copy(marshaledMetadata[MARSHALED_HASH_START:MARSHALED_HASH_END], metadata.hash.CastToBytes())

	marshaledReceivedTime, err := metadata.receivedTime.MarshalBinary()
	if err != nil {
		return nil, ErrMarshallFailed.Derive(err, "failed to marshal received time")
	}
	copy(marshaledMetadata[MARSHALED_RECEIVED_TIME_START:MARSHALED_RECEIVED_TIME_END], marshaledReceivedTime)

	var booleanFlags bitutils.BitMask
	if metadata.solid {
		booleanFlags = booleanFlags.SetFlag(0)
	}
	if metadata.liked {
		booleanFlags = booleanFlags.SetFlag(1)
	}
	if metadata.finalized {
		booleanFlags = booleanFlags.SetFlag(2)
	}
	marshaledMetadata[MARSHALED_FLAGS_START] = byte(booleanFlags)

	return marshaledMetadata, nil
}

func (metadata *TransactionMetadata) Unmarshal(data []byte) errors.IdentifiableError {
	metadata.hashMutex.Lock()
	defer metadata.hashMutex.Unlock()
	metadata.receivedTimeMutex.Lock()
	defer metadata.receivedTimeMutex.Unlock()
	metadata.solidMutex.Lock()
	defer metadata.solidMutex.Unlock()
	metadata.likedMutex.Lock()
	defer metadata.likedMutex.Unlock()
	metadata.finalizedMutex.Lock()
	defer metadata.finalizedMutex.Unlock()

	metadata.hash = ternary.Trinary(typeutils.BytesToString(data[MARSHALED_HASH_START:MARSHALED_HASH_END]))

	if err := metadata.receivedTime.UnmarshalBinary(data[MARSHALED_RECEIVED_TIME_START:MARSHALED_RECEIVED_TIME_END]); err != nil {
		return ErrUnmarshalFailed.Derive(err, "could not unmarshal the received time")
	}

	booleanFlags := bitutils.BitMask(data[MARSHALED_FLAGS_START])
	if booleanFlags.HasFlag(0) {
		metadata.solid = true
	}
	if booleanFlags.HasFlag(1) {
		metadata.liked = true
	}
	if booleanFlags.HasFlag(2) {
		metadata.finalized = true
	}

	return nil
}

// endregion ///////////////////////////////////////////////////////////////////////////////////////////////////////////

// region database functions ///////////////////////////////////////////////////////////////////////////////////////////

// endregion ///////////////////////////////////////////////////////////////////////////////////////////////////////////

// region constants and variables //////////////////////////////////////////////////////////////////////////////////////

const (
	MARSHALED_HASH_START          = 0
	MARSHALED_RECEIVED_TIME_START = MARSHALED_HASH_END
	MARSHALED_FLAGS_START         = MARSHALED_RECEIVED_TIME_END

	MARSHALED_HASH_END          = MARSHALED_HASH_START + MARSHALED_HASH_SIZE
	MARSHALED_RECEIVED_TIME_END = MARSHALED_RECEIVED_TIME_START + MARSHALED_RECEIVED_TIME_SIZE
	MARSHALED_FLAGS_END         = MARSHALED_FLAGS_START + MARSHALED_FLAGS_SIZE

	MARSHALED_HASH_SIZE          = 81
	MARSHALED_RECEIVED_TIME_SIZE = 15
	MARSHALED_FLAGS_SIZE         = 1

	MARSHALED_TOTAL_SIZE = MARSHALED_FLAGS_END
)

// endregion ////////////////////////////////////////////////////////////////////////////////////////////////////////////
